
package br.ufsc.das.paxosatwar.roles;

import br.ufsc.das.paxosatwar.messages.FreezeProof;
import br.ufsc.das.paxosatwar.messages.Proof;
import br.ufsc.das.paxosatwar.messages.RequestAutenticator;
import br.ufsc.das.paxosatwar.roles.util.LeaderModule;
import br.ufsc.das.tom.usage.TOMLayer;
import br.ufsc.das.tom.util.ProposeValue;
import br.ufsc.das.tom.timer.RTCollect;
import br.ufsc.das.util.CommunicationSystem;

import br.ufsc.das.paxosatwar.roles.util.ExecutionManager;
import br.ufsc.das.paxosatwar.roles.util.Execution;
import br.ufsc.das.paxosatwar.roles.util.ProofVerifier;
import br.ufsc.das.paxosatwar.roles.util.Round;
import br.ufsc.das.paxosatwar.roles.util.TimeoutTask;
import br.ufsc.das.paxosatwar.messages.CollectProof;
import br.ufsc.das.paxosatwar.messages.PaxosMessage;
import br.ufsc.das.paxosatwar.messages.MessageFactory;
import br.ufsc.das.paxosatwar.Consensus;
import br.ufsc.das.util.Configuration;
import br.ufsc.das.util.TOMConfiguration;
import java.security.SignedObject;

import java.util.concurrent.Semaphore;
import java.util.Timer;
import java.util.concurrent.locks.ReentrantLock;

/**
 * This class represents the acceptor role in the paxos protocol.
 * This class work together with the TOMulticastLayer class in order to
 * supply a atomic multicast service.
 */

public class Acceptor{
    
    private Timer timer = new Timer();
    private int me;
    private ExecutionManager manager;
    private MessageFactory factory;
    private ProofVerifier verifier;
    private CommunicationSystem communication;
    private LeaderModule leaderModule;
    private TOMLayer multicastLayer;
    private AcceptedPropose nextProp = null;
    private Semaphore mutex = new Semaphore(1);
    private TOMConfiguration conf;
    
    /**
     * Creates a new instance of Acceptor.
     */
    public Acceptor(CommunicationSystem communication, MessageFactory factory, 
            ProofVerifier verifier, LeaderModule lm, TOMConfiguration conf) {
        this.communication = communication;
        this.communication.setAcceptor(this);
        this.me = conf.getProcessId();
        this.factory = factory;
        this.verifier = verifier;
        this.leaderModule = lm;
        this.conf = conf;
    }
    
    public boolean verifySignature(SignedObject so, int sender){
        return this.verifier.validSignature(so,sender);
    }
    
    public SignedObject sign(RTCollect rtc){
        return this.verifier.sign(rtc);
    }
    
    public void setManager(ExecutionManager manager) {
        this.manager = manager;
    }
    
    public void setTOMLayer(TOMLayer tom){
        this.multicastLayer = tom;
    }
    
    public void getMEZone(){
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
    }
    
    public void leaveMEZone(){
        this.mutex.release();
    }
    
    
    //called by Execution when a new round is created
    public void scheduleTimeout(Round round) {
        TimeoutTask task = new TimeoutTask(this, round);
        round.setTimeoutTask(task);
        timer.schedule(task, round.getTimeout());
    }
    
    //called by timertasks associated with rounds
    public void timeout(Round round) {
        //br.ufsc.das.util.Logger.println(round.getExecution().getId()+" timeout o round "+round.getNumber()+ " do pedido " + round.getExecution().getId());
        if(!round.getExecution().isDecided() && !round.isFrozen() && !round.isRemoved()) {
            //br.ufsc.das.util.Logger.println(round.getExecution().getId()+" congelou o round "+round.getNumber()+ " do pedido " + round.getExecution().getId());
            round.freeze();
            communication.send(manager.getOtherAcceptors(),
                    factory.createFreeze(round.getExecution().getId(),round.getNumber()));
            computeFreeze(round.getExecution().getId(),round);
        }
    }
    
    //called by communication layer to delivery paxos messages
    public void deliver(PaxosMessage msg) {
        this.getMEZone();
        if(manager.checkLimits(msg,conf)){
            if(msg.getPaxosType() == MessageFactory.PROPOSE){
                proposeReceived(msg);
            }else{
                processMessage(msg, manager.getExecution(msg.getNumber()));
            }
        }
        this.leaveMEZone();
    }
    
    public void processMessage(PaxosMessage msg, Execution execution) {
        Round round = execution.getRound(msg.getRound());
        switch(msg.getPaxosType()){
            case MessageFactory.WEAK:{
                weakAcceptReceived(msg.getNumber(),round,msg.getFrom(), msg.getValue());
            }break;
            
            case MessageFactory.STRONG:{
                strongAcceptReceived(msg.getNumber(),round,msg.getFrom(),msg.getValue());
            }break;
            
            case MessageFactory.DECIDE:{
                decideReceived(msg.getNumber(),round,msg.getFrom(),msg.getValue());
            }break;
            
            case MessageFactory.FREEZE:
                freezeReceived(msg.getNumber(),round,msg.getFrom());
        }
    }
    
    
    public void proposeReceived(PaxosMessage msg) {
        Object value = msg.getValue();
        int p = msg.getFrom();
        int eid = msg.getNumber();
        Proof proof = (Proof)msg.getProof();
        if(msg.getRound() == 0){
            //br.ufsc.das.util.Logger.println("Recebendo proposta num " + eid + ", round=0 do processo: "+p);
            if(leaderModule.getLeader(eid,msg.getRound()) == p){
                executePropose(eid,msg.getRound(),value,proof);
            }
        }else{
            if(proof != null){
                ////br.ufsc.das.util.Logger.println("Recebendo proposta para r!=0 do processo: "+p);
                //br.ufsc.das.util.Logger.println("Recebendo proposta num " + eid + ", round!=0 do processo: "+p);
                CollectProof[] collected = verifier.checkValid(eid, msg.getRound()-1, proof.getProofs());
                if(verifier.isTheLeader(p,collected)){
                    leaderModule.addLeaderInfo(eid,msg.getRound(),p);
                    checkRequestsDiscarts(eid, collected);
                    if(value != null && (verifier.good(value,collected,true))){
                        executePropose(eid,msg.getRound(),value,proof);
                    }else{
                        if(checkConsensusDiscart(eid, collected,true)){
                            leaderModule.addLeaderInfo(eid,0,p);
                        }
                    }
                    if(proof.getNextPropose() != null && verifier.good(proof.getNextPropose(),collected,false)){
                        int round = verifier.getNextExecRound(collected);
                        if(multicastLayer.getInExec() == eid+1){
                            manager.getExecution(eid+1).removeRounds(round-1);
                            executePropose(eid+1,round,value,proof);
                        }else{
                            nextProp = new AcceptedPropose(eid+1,round,value,proof);
                        }
                    }else{
                        if(checkConsensusDiscart(eid+1, collected,false)){
                            leaderModule.addLeaderInfo(eid+1,0,p);
                        }
                    }
                }
            }
        }
    }
    
    public boolean executeAcceptedPendent(int eid){
        if(nextProp != null && nextProp.eid == eid){
            executePropose(nextProp.eid,nextProp.r,nextProp.obj,nextProp.p);
            nextProp = null;
            return true;
        }else{
            nextProp = null;
            return false;
        }
    }
    
    private void executePropose(int eid, int r, Object value, Proof proof){
        //br.ufsc.das.util.Logger.println("(Acceptor) (" + eid + ") executePropose ");
        Execution execution = manager.getExecution(eid);
        Round round = execution.getRound(r);
        round.propValue = value;
        if(eid == multicastLayer.getLastExec()+1){
            multicastLayer.setInExec(eid);
        }
       
        if((proof != null || multicastLayer.containRequests(round)) && !round.isWeakSetted(me)){
            //br.ufsc.das.util.Logger.println("(Acceptor) (" + eid + ") executePropose a enviar WEAK com value = "+value);
            round.setWeak(me,value);
          
            /*if(r==0 && leaderModule.getLeader(eid,r) == 0){
                return;
            }*/
            
            
            int[] otherProcesses = manager.getOtherAcceptors();
            communication.send(otherProcesses,factory.createWeak(eid, r, value));
            computeWeak(eid,round,value);
        }
        else if (!multicastLayer.containRequests(round)) {
           //br.ufsc.das.util.Logger.println("(Acceptor) (" + eid + ") executePropose nao enviou WEAK pq nao ha requests deste round ");
        }
        else {
           round.setWeak(me,value);
           //br.ufsc.das.util.Logger.println("(Acceptor) (" + eid + ") executePropose nao enviou WEAK pq ja foi enviado um WEAK ");
        }
    }
    
    private void checkRequestsDiscarts(int eid, CollectProof[] proof){
        // s� pode descartar reqs do consenso q foi congelado...
        if(multicastLayer.getLastExec() < eid){
            ProposeValue disc = verifier.getRequests(proof,false,true);
            for(int i = 0; i < disc.getValues().length; i++){
                multicastLayer.discartRequest(disc.getValues()[i]);
            }
            multicastLayer.checkTimer();
        }
    }
    
    
    /*
     * retorna true se � pra atualizar o lider, false caso contrario...
     */
    private boolean checkConsensusDiscart(int eid, CollectProof[] proof, boolean in){
        if(multicastLayer.getLastExec() < eid){
            if(verifier.getGoodValue(proof,in) == null){
                //br.ufsc.das.util.Logger.println("Descartando o consenso "+eid);
                if(multicastLayer.getInExec() == eid){
                    multicastLayer.setInExec(-1);
                }
                Execution exec = manager.removeExecution(eid);
                if(exec != null){
                    exec.removeRounds(-1);//cancela os timeouts dos rounds
                }
                if(multicastLayer.getLastExec()+1 == eid){
                    return true;
                }
            }
        }
        return false;
    }
    
    public void weakAccept(Round round){
       
        if(!round.isWeakSetted(me)){
            round.setWeak(me,round.propValue);
           
            /*if(round.getNumber() ==0 &&
                    leaderModule.getLeader(round.getExecution().getId(),round.getNumber()) == 0){
                return;
            }*/
            
            //br.ufsc.das.util.Logger.println("(Acceptor) (weakAccept)");
            
            
            int[] otherProcesses = manager.getOtherAcceptors();
            communication.send(otherProcesses,
                    factory.createWeak(round.getExecution().getId(),round.getNumber(),round.propValue));
            computeWeak(round.getExecution().getId(),round,round.propValue);
        }
        else {
           
        }
    }
    
    private void weakAcceptReceived(int eid, Round round, int a, Object value) {
    
        //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Recebeu weak de "+a );
        round.setWeak(a,value);
        this.multicastLayer.computeAck(eid-1,a);
        ////br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Weak de "+a + "(depois do computeAck)");
        ////br.ufsc.das.util.Logger.println("round "+round);
        if(!round.isWeakSetted(me)){
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Eh o primeiro weak q processo, vem de "+a);
            int weakAccepted = round.countWeak(value);
            if(weakAccepted > manager.quorumF){
                if(eid == multicastLayer.getLastExec()+1){
                    multicastLayer.setInExec(eid);
                }
                 round.setWeak(me,value);
                /*if(round.getNumber() ==0 &&
                        leaderModule.getLeader(round.getExecution().getId(),round.getNumber()) == 0){
                    return;
                }*/
                
              
               
                int[] otherProcesses = manager.getOtherAcceptors();
                communication.send(otherProcesses,
                        factory.createWeak(eid, round.getNumber(), value));
            }
        }
        else {
             //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Ja enviei o meu WEAK anteriormente");
            
        }
        computeWeak(eid,round,value);
    }
    
    private void computeWeak(int eid, Round round, Object value) {
      
        int weakAccepted = round.countWeak(value);
        //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") tenho " + weakAccepted + " weaks ");
        if(weakAccepted > manager.quorumStrong) {
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") tenho weaks suficientes para enviar um STRONG ");
            if(!round.isStrongSetted(me)) {
                //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Vou enviar STRONG ");
                round.setStrong(me, value);
                
                int[] otherProcesses = manager.getOtherAcceptors();
                communication.send(otherProcesses,
                        factory.createStrong(eid,round.getNumber(), value));
                computeStrong(eid,round,value);
            }
            else {
             //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Ja enviei o meu STRONG anteriormente");
             
            }
            
            if(weakAccepted > manager.quorumFastDecide && !round.getExecution().isDecided()) {
                
                round.setDecide(me, round.propValue);
               
                int[] otherProcesses = manager.getOtherAcceptors();
                communication.send(otherProcesses,
                        factory.createDecide(eid, round.getNumber(),round.propValue));
                
                //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Vou decidir pq recebi weaks suficientes");
                decide(round,value);                
            }
            else {
                
            }
        }
        else {
           
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Preciso de um total de " + (manager.quorumStrong+1) + " weaks ");
        }
    }
    
    private void strongAcceptReceived(int eid, Round round, int a, Object value) {
        round.setStrong(a,value);
        computeStrong(eid,round,value);
    }
    
    private void computeStrong(int eid, Round round, Object value) {
      
        if(round.countStrong(value) > manager.quorum2F && !round.getExecution().isDecided()) {
            round.setDecide(me, value);
           
            int[] otherProcesses = manager.getOtherAcceptors();
            communication.send(otherProcesses,
                    factory.createDecide(eid,round.getNumber(),round.propValue));
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Vou decidir pq recebi strongs suficientes");
            decide(round,value);
            //decide(round,round.propValue); pjsousa: comentei pq estava a passar nulls de vez em quando
        }
        else if (!round.getExecution().isDecided()) {
          
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") computeStrong: Recebi "+round.countStrong(value)+" STRONGS mas preciso de receber mais do que "+manager.quorum2F);
        }
        else {
           
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") computeStrong: Já há uma decisão");
        }
    }
    
    private void decideReceived(int eid, Round round, int a, Object value) {
        
        //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") a processar DECIDE de " + a);
        round.setDecide(a,value);
        int numDecides = round.countDecide(value);
        if(numDecides > manager.quorumF &&
                !round.getExecution().isDecided()) {
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") recebi mais do que " + manager.quorumF + " decides ");
            round.setDecide(me, value);
           
            int[] otherProcesses = manager.getOtherAcceptors();
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") A enviar DECIDE para os outros ");
            communication.send(otherProcesses,
                    factory.createDecide(eid,round.getNumber(),value));
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Vou decidir pq recebi DECIDEs suficentes");
            decide(round,value);
        }
        else if (round.getExecution().isDecided()){
            
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") já recebi DECIDEs suficientes para decidir, mas já decidi");
        }
        else {
            //br.ufsc.das.util.Logger.println("(Acceptor) ("+eid+") Ainda nao recebi DECIDEs suficientes para decidir. Recebi "+numDecides+ " e preciso de receber mais do que "+manager.quorumF);
        }
    }
    
    private void freezeReceived(int eid, Round round, int a) {
        round.addFreeze(a);
        if(round.countFreeze() > manager.quorumF && !round.isFrozen()) {
            round.freeze();
            communication.send(manager.getOtherAcceptors(),
                    factory.createFreeze(eid,round.getNumber()));
        }
        computeFreeze(eid,round);
    }
    
    private void computeFreeze(int eid, Round round){
        if(round.countFreeze() > manager.quorum2F && !round.isCollected()) {
            round.collect();
            round.getTimeoutTask().cancel();
            Round nextRound = round.getExecution().getRound(round.getNumber()+1);
            if(!nextRound.isWeakSetted(me)){ //n�o esta executando um round maior
                Execution exec = manager.getExecution(eid);
                int newLeader = (leaderModule.getLeader(eid,round.getNumber())+1)%conf.getN();
                leaderModule.addLeaderInfo(eid,round.getNumber()+1,newLeader);
                if(exec.isDecided()){
                    Round r = exec.getDecidedRound();
                    if(!(r.getNumber() > round.getNumber())){
                        FreezeProof fpIn = createProof(eid,round);
                        Execution nextExec = manager.isInExecution(eid+1);
                        CollectProof clProof = null;
                        if(nextExec != null){
                            Round last = nextExec.getLastRound();
                            last.freeze();
                            clProof = new CollectProof(fpIn,createProof(eid+1,last),newLeader);
                        }else{
                            clProof = new CollectProof(fpIn,null,newLeader);
                        }
                        communication.send(new int[]{newLeader},
                                factory.createCollect(eid, round.getNumber(),verifier.sign(clProof)));
                    }
                }else{
                    CollectProof clProof = new CollectProof(createProof(eid,round),null,newLeader);
                    communication.send(new int[]{newLeader},
                            factory.createCollect(eid, round.getNumber(),verifier.sign(clProof)));
                }
            }
        }
    }
    
    private FreezeProof createProof(int eid, Round r){
        return new FreezeProof(me, eid, r.getNumber(), r.getWeak(me), r.getStrong(me),
                r.getDecide(me),getAutenticators(r));
    }
    
    private RequestAutenticator[] getAutenticators(Round r){
        RequestAutenticator[] ret = null;
        if(r.propValue != null){
            byte[][] prop = ((ProposeValue) r.propValue).getValues();
            ret = new RequestAutenticator[prop.length];
            for(int i = 0; i < ret.length; i++){
                ret[i] = new RequestAutenticator(prop[i],
                        multicastLayer.containRequest(prop[i]));
            }
        }
        return ret;
    }
    
    private void decide(Round round, Object value) {
        if(value != null){
           
            leaderModule.decided(round.getExecution().getId(),
                    leaderModule.getLeader(round.getExecution().getId(), round.getNumber()));
            round.getTimeoutTask().cancel();
            round.getExecution().decide(value,round.getNumber());
           
        }
        else {
            //br.ufsc.das.util.Logger.println("(Acceptor) decide recebeu valor NULL, logo não houve decisão de facto");
        }
    }
    
    private class AcceptedPropose{
        public int eid;
        public int r;
        public Object obj;
        public Proof p;
        
        public AcceptedPropose(int eid, int r, Object obj, Proof p){
            this.eid = eid;
            this.r = r;
            this.obj = obj;
            this.p = p;
        }
    }
}
