
package br.ufsc.das.paxosatwar.roles;
import br.ufsc.das.paxosatwar.messages.Proof;
import br.ufsc.das.util.CommunicationSystem;
import br.ufsc.das.paxosatwar.roles.util.Execution;
import br.ufsc.das.paxosatwar.roles.util.ExecutionManager;
import br.ufsc.das.paxosatwar.roles.util.ProofVerifier;
import br.ufsc.das.paxosatwar.roles.util.Round;
import br.ufsc.das.paxosatwar.messages.CollectProof;
import br.ufsc.das.paxosatwar.messages.PaxosMessage;
import br.ufsc.das.paxosatwar.messages.MessageFactory;
import static br.ufsc.das.paxosatwar.messages.MessageFactory.*;
import br.ufsc.das.paxosatwar.Consensus;
import br.ufsc.das.util.TOMConfiguration;
import java.security.SignedObject;


/**
 * This class represents the proposer role in the paxos protocol.
 **/

public class Proposer {
    
    private ExecutionManager manager = null;
    private MessageFactory factory;
    private ProofVerifier verifier;
    
    private CommunicationSystem communication;
    
    private TOMConfiguration conf;
    
    /**
     * Creates a new instance of Proposer
     */
    public Proposer(CommunicationSystem communication, MessageFactory factory,
            ProofVerifier verifier, TOMConfiguration conf) {
        this.communication = communication;
        this.communication.setProposer(this);
        this.verifier = verifier;
        this.factory = factory;
        this.conf = conf;
    }
    
    public void setManager(ExecutionManager manager) {
        this.manager = manager;
    }

    /** This method is called by the TOMLayer (or any other)
     * to start the execution of one instance of the paxos protocol.
     */
    public void startExecution(int eid, Object value) {
        Execution execution = manager.getExecution(eid);
        Round round = execution.getRound(0);
        if(!round.proposed) {
            round.proposed = true;
            //br.ufsc.das.util.Logger.println("Enviando proposta sem provas para consenso "+eid);
            communication.send(manager.getAcceptors(),
                    factory.createPropose(eid, 0, value, null));
        }
        else {
            //br.ufsc.das.util.Logger.println("(Proposer) Proposta já foi enviada anteriormente");
        }
    }
    
    
    public void deliver(PaxosMessage msg) {
        if(manager.checkLimits(msg,conf)){
            int msgType = msg.getPaxosType();
            if(msgType == COLLECT) {
                collectReceived(msg.getNumber(), msg.getRound(),
                        msg.getFrom(), (SignedObject)msg.getProof());
            } else if((manager.getAcceptor() == null) &&
                    (msgType == WEAK || msgType == STRONG || msgType == DECIDE)) {
                paxosMessageReceived(msg.getNumber(), msg.getRound(),
                        msgType, msg.getFrom(), msg.getValue());
            }
        }
    }
    
    /**
     * This method is executed when a COLLECT message is received.
     */
    private void collectReceived(int eid, int r, int a, SignedObject proof) {
        if(proof != null && verifier.validSignature(proof,a)){
            CollectProof cp = null;
            try{
                cp = (CollectProof) proof.getObject();
            }catch(Exception e){
                e.printStackTrace(System.out);
            }
            if((cp != null) && (cp.getProofs(true) != null) &&
                    // a prova para o round q esta sendo congelado (FreezeProof[0]) precisa ser v�lida
                    verifier.validProof(eid,r,cp.getProofs(true)) &&
                    // este processo � o novo l�der
                    (cp.getLeader() == conf.getProcessId())){
                
                Execution execution = manager.getExecution(eid);
                Round round = execution.getRound(r+1);
                round.collectProof(proof,a,conf);
                if(!round.proposed &&
                        verifier.countProofs(round.proofs) > manager.quorumStrong) {
                    round.proposed = true;
                    Object inProp = verifier.getGoodValue(round.proofs,true);
                    Object nextProp = verifier.getGoodValue(round.proofs,false);
                    
                    ////br.ufsc.das.util.Logger.println("Enviando proposta com provas");
                    communication.send(manager.getAcceptors(),
                            factory.createPropose(eid, r+1, inProp,new Proof(round.proofs,nextProp)));
                }
            }
        }
    }
    
    private void paxosMessageReceived(int eid, int rid, int msgType,
            int sender, Object value) {
        Round round = manager.getExecution(eid).getRound(rid);
        if(msgType == WEAK) {
            round.setWeak(sender, value);
            if(round.countWeak(value) > manager.quorumFastDecide) {
                manager.getExecution(eid).decide(value,round.getNumber());
            }
        } else if(msgType == STRONG) {
            round.setStrong(sender, value);
            if(round.countStrong(value) > manager.quorum2F) {
                manager.getExecution(eid).decide(value,round.getNumber());
            }
        } else if(msgType == DECIDE) {
            round.setDecide(sender, value);
            if(round.countDecide(value) > manager.quorumF) {
                manager.getExecution(eid).decide(value,round.getNumber());
            }
        }
    }
}
