package br.ufsc.das.paxosatwar.roles.util;

import br.ufsc.das.paxosatwar.messages.FreezeProof;
import br.ufsc.das.paxosatwar.messages.RequestAutenticator;
import br.ufsc.das.tom.util.ProposeValue;
import br.ufsc.das.tom.timer.RTCollect;
import br.ufsc.das.util.TOMUtil;
import br.ufsc.das.util.TOMConfiguration;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.SignedObject;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import br.ufsc.das.paxosatwar.messages.CollectProof;
import java.util.LinkedList;
import java.util.List;

/**
 * This class is used to process data relacioned with freezed rounds. 
 * Generate proposes - good values, verify the proposed values and so on...
 */

public class ProofVerifier {
    
    private int quorumF;
    private int quorumStrong;
    
    private PublicKey[] publickeys;
    private PrivateKey prk = null;
    private Signature engine;
    
    /** Creates a new instance of ProofVerifier */
    public ProofVerifier(TOMConfiguration conf) {
        this.quorumF = conf.getF();
        this.quorumStrong = (int)((conf.getN()+quorumF)/2);
        this.publickeys = conf.getRSAPublicKeys();
        this.prk = conf.getRSAPrivateKey();
        try{
            this.engine = Signature.getInstance("SHA1withRSA");
        }catch(Exception e){
            e.printStackTrace();
        }
    }
    
    public SignedObject sign(CollectProof cp){
        try{
            return new SignedObject(cp,prk,engine);
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }
    
    public SignedObject sign(RTCollect trc){
        try{
            return new SignedObject(trc,prk,engine);
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }
    
    public int countProofs(SignedObject[] proofs) {
        int validProofs = 0;
        for(int i = 0; i < proofs.length; i++) {
            if(proofs[i] != null) {
                validProofs++;
            }
        }
        return validProofs;
    }
    
    public Object getGoodValue(SignedObject[] proofs, boolean in) {
        try{
            CollectProof[] cps = new CollectProof[proofs.length];
            for(int i = 0; i < proofs.length; i++){
                if(proofs[i] != null){
                    cps[i] = (CollectProof) proofs[i].getObject();
                }
            }
            return getGoodValue(cps,in);
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }
    
    public Object getGoodValue(CollectProof[] proofs, boolean in) {
        LinkedList<ProposeValue> poss = buildPoss(proofs,in);
        LinkedList<ProposeValue> acc = buildAcc(proofs,in);
        for(Iterator i = acc.iterator(); i.hasNext(); ) {
            Object value = i.next();
            if(poss.contains(value)) {
                return value;
            }
        }
        
        
        ProposeValue reqs = getRequests(proofs,true,in);
        if(reqs.getValues().length > 0){
            return reqs;
        }
        return null;
    }
    
    /**
     * Called by acceptors to verify if some proposal is good.
     *
     */
    public boolean good(Object value, CollectProof[] proofs, boolean  in) {
        
        LinkedList<ProposeValue> poss = buildPoss(proofs,in);
        LinkedList<ProposeValue> acc = buildAcc(proofs,in);
        
        //condition G2
        if(acc.contains(value) && (poss.contains(value) || poss.isEmpty())){
            return true;
        }
        
        //condition G1
        if(poss.isEmpty()){
            ProposeValue prop = getRequests(proofs,true,in);
            if(prop.equals(value)){
                return true;
            }
        }
        return false;
    }
    
    
    private boolean contain(RequestAutenticator ra, RequestAutenticator[] others){
        for(int i = 0; i < others.length; i++){
            if(TOMUtil.equalsHash(others[i].req,ra.req) &&
                    others[i].contain == ra.contain){
                return true;
            }
        }
        return false;
    }
    
    public ProposeValue getRequests(CollectProof[] proofs, boolean auth, boolean in){
        List l = new LinkedList();
        RequestAutenticator[] ra = null;
        RequestAutenticator[] comp = null;
        for(int i=0; i < proofs.length; i++) {
            ra = null;
            if(proofs[i] != null){
                ra = proofs[i].getAutenticators(in);
            }
            if(ra != null){
                for(int x = 0; x < ra.length; x++){
                    int count = 1;
                    if(ra[x].contain == auth){
                        for(int j=i+1; j < proofs.length; j++) {
                            comp = null;
                            if(proofs[j] != null){
                                comp = proofs[j].getAutenticators(in);
                            }
                            if(comp != null && contain(ra[x],comp)){
                                count++;
                            }
                        }
                    }
                    if(count > quorumF){
                        boolean add = true;
                        for(int z = 0; z < l.size() && add ; z++){
                            if(TOMUtil.equalsHash(ra[x].req,(byte[])l.get(z))){
                                add = false;
                            }
                        }
                        if(add){
                            l.add(ra[x].req);
                        }
                    }
                    
                }
            }
        }
        
        byte[][] reqs = new byte[l.size()][0];
        for(int i = 0; i < l.size(); i++){
            reqs[i] = (byte[])l.get(i);
        }
        
        return new ProposeValue(reqs);
    }
    
    
    
    
    public int getNextExecRound(CollectProof[] proof){
        for(int i = 0; i < proof.length; i++){
            if(proof[i].getProofs(false) != null){
                int r = proof[i].getProofs(false).getRound();
                int c = 1;
                for(int j = i+1; j < proof.length; j++){
                    if(proof[j].getProofs(false) != null){
                        if(r == proof[j].getProofs(false).getRound()){
                            c++;
                        }
                    }
                }
                if(c > quorumF){
                    return r;
                }
            }
        }
        return -1;
    }
    
    
    public boolean validProof(int eid, int round, FreezeProof proof) {
        return (proof != null) && (proof.getEid() == eid) && (proof.getRound() == round);
    }
    
    public CollectProof[] checkValid(int eid, int round, SignedObject[] proof){
        if(proof == null){
            return null;
        }
        Collection<CollectProof> valid = new HashSet<CollectProof>();
        try{
            for(int i = 0; i < proof.length; i++){
                if(proof[i] != null && validSignature(proof[i],i)){
                    CollectProof cp = (CollectProof)proof[i].getObject();
                    if(validProof(eid,round,cp.getProofs(true))){
                        valid.add(cp);
                    }
                }
            }
        }catch(Exception e){
            e.printStackTrace();
        }
        return valid.toArray(new CollectProof[0]);
    }
    
    
    public boolean validSignature(SignedObject so, int sender) {
        try{
            return so.verify(this.publickeys[sender],engine);
        }catch(Exception e){
            e.printStackTrace();
        }
        return false;
    }
    
    public boolean isTheLeader(int l, CollectProof[] proof){
        int c = 0;
        for(int i = 0; i < proof.length; i++){
            if(proof[i] != null && proof[i].getLeader() == l){
                c++;
            }
        }
        return c > quorumF;
    }
    
    private LinkedList<ProposeValue> buildPoss(CollectProof[] proofs, boolean in) {
        LinkedList<ProposeValue> poss = new LinkedList<ProposeValue>();
        
        for(int i=0; i < proofs.length; i++) {
            ProposeValue w = null;
            if(proofs[i] != null &&
                    proofs[i].getProofs(in) != null){
                
                w = (ProposeValue)proofs[i].getProofs(in).getWeak();
            }
            if(w != null) {
                int countW = 0, countS = 0;
                for(int j=0; j < proofs.length; j++) {
                    if(proofs[j] != null && proofs[j].getProofs(in) != null){
                        if(w.equals(proofs[j].getProofs(in).getWeak())){
                            countW++;
                        }
                        if(w.equals(proofs[j].getProofs(in).getStrong())){
                            countS++;
                        }
                    }
                }
                if((countW > quorumStrong || countS > quorumF) && !poss.contains(w)) {
                    poss.add(w);
                }
            }
        }
        return poss;
    }
    
    private LinkedList<ProposeValue> buildAcc(CollectProof[] proofs, boolean in) {
        LinkedList<ProposeValue> acc = new LinkedList<ProposeValue>();
        
        for(int i=0; i < proofs.length; i++) {
            ProposeValue w = null;
            if(proofs[i] != null &&
                    proofs[i].getProofs(in) != null){
                
                w = (ProposeValue)proofs[i].getProofs(in).getWeak();
            }
            if(w != null) {
                int count = 0;
                
                for(int j=0; j < proofs.length; j++) {
                    
                    if(proofs[j] != null &&
                            proofs[j].getProofs(in) != null && 
                                w.equals(proofs[j].getProofs(in).getWeak())){
                        count++;
                    }
                }
                if(count > quorumF && !acc.contains(w)) {
                    acc.add(w);
                }
            }
        }
        
        return acc;
    }
}
