

package br.ufsc.das.paxosatwar.roles.util;

import br.ufsc.das.util.Configuration;
import br.ufsc.das.util.TOMConfiguration;
import java.io.Serializable;
import java.security.SignedObject;
import java.util.Arrays;
import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

import br.ufsc.das.paxosatwar.messages.CollectProof;
import java.util.concurrent.locks.ReentrantLock;


public class Round implements Serializable {
    
    private transient Execution execution;
    private transient TimeoutTask timeoutTask;
    
    private int number;
    private int me;
    
    private boolean[] weakSetted;
    private boolean[] strongSetted;
    private boolean[] decideSetted;
    
    private Object[] weak;
    private Object[] strong;
    private Object[] decide;
    
    public boolean proposed;
    public Object propValue;
    public SignedObject[] proofs;
    
    private int leader;
    
    private Collection<Integer> freeze = null;
    
    private boolean frozen = false;
    private boolean collected = false;
    
    private long timeout;
    
    private LeaderModule leaderModule;
    
    private boolean alreadyRemoved = false;
    
    private ReentrantLock lockWeak;
    private ReentrantLock lockStrong;
    private ReentrantLock lockDecide;
    
    
    /** Creates a new instance of Round for acceptors */
    protected Round(Execution parent, int number, long timeout) {
        this.execution = parent;
        this.number = number;
        
        ExecutionManager manager = execution.getManager();
        
        this.me = manager.getProcessId();
        
        int[] acceptors = manager.getAcceptors();
        int n = acceptors.length;
        
        this.weakSetted = new boolean[n];
        this.strongSetted = new boolean[n];
        this.decideSetted = new boolean[n];
        
        this.weak = new Object[n];
        this.strong = new Object[n];
        this.decide = new Object[n];
        
        this.lockWeak = new ReentrantLock();
        this.lockStrong = new ReentrantLock();
        this.lockDecide = new ReentrantLock();
        
        if(number > 0) {
            Round previousRound =  execution.getRound(number-1);
            System.arraycopy(previousRound.getWeak(), 0, weak, 0, n);
            System.arraycopy(previousRound.getStrong(), 0, strong, 0, n);
            System.arraycopy(previousRound.getDecide(), 0, decide, 0, n);
            this.timeout = previousRound.getTimeout()*2;
        }else{
            this.timeout = timeout;
        }
        
        if(manager.getAcceptor() != null) {
            manager.getAcceptor().scheduleTimeout(this);
        }
    }
    
    public void setRemoved(){
        this.alreadyRemoved = true;
    }
    
    public boolean isRemoved(){
        return this.alreadyRemoved;
    }
    
    public void collectProof(SignedObject p, int a, TOMConfiguration conf){
        if(proofs == null) {
            proofs = new SignedObject[conf.getN()];
        }
        proofs[a] = p;
    }
    
    public long getTimeout(){
        return this.timeout;
    }
    
    
    public int getNumber() {
        return number;
    }
    
    public Execution getExecution() {
        return execution;
    }
    
    public void setTimeoutTask(TimeoutTask timeoutTask) {
        this.timeoutTask = timeoutTask;
    }
    
    public TimeoutTask getTimeoutTask() {
        return timeoutTask;
    }
    
    /**
     * Getter for property leader.
     * @return Value of the property.
     */
    public int getLeader() {
        return leader;
    }
    
    public boolean isWeakSetted(int acceptor) {
        return weakSetted[acceptor];
    }
    
    public boolean isStrongSetted(int acceptor) {
        return strongSetted[acceptor];
    }
    
    public boolean isDecideSetted(int acceptor) {
        return decideSetted[acceptor];
    }
    
    /**
     * Indexed getter for property weak.
     * @param acceptor Index of the property.
     * @return Value of the property at <CODE>index</CODE>.
     */
    public Object getWeak(int acceptor) {
        return this.weak[acceptor];
    }
    
    /**
     * Getter for property weak.
     * @return Value of property weak.
     */
    public Object[] getWeak() {
        return this.weak;
    }
    
    /**
     * Indexed setter for property weak.
     * @param acceptor Index of the property.
     * @param value New value of the property at <CODE>index</CODE>.
     */
    public void setWeak(int acceptor, Object value) {
         lockWeak.lock();
            if(!weakSetted[acceptor] && !isFrozen()) { //it can only be setted one time
                weak[acceptor] = value;
                weakSetted[acceptor] = true;
            }
         lockWeak.unlock();
    }
    
    /**
     * Indexed getter for property strong.
     * @param acceptor Index of the property.
     * @return Value of the property at <CODE>index</CODE>.
     */
    public Object getStrong(int acceptor) {
        return strong[acceptor];
    }
    
    /**
     * Getter for property strong.
     * @return Value of property strong.
     */
    public Object[] getStrong() {
        return strong;
    }
    
    /**
     * Indexed setter for property strong.
     * @param acceptor Index of the property.
     * @param value New value of the property at <CODE>index</CODE>.
     */
    public void setStrong(int acceptor, Object value) {
         lockStrong.lock();
            if(!strongSetted[acceptor] && !isFrozen()) { //it can only be setted one time
                strong[acceptor] = value;
                strongSetted[acceptor] = true;
            }
         lockStrong.unlock();
    }
    
    /**
     * Indexed getter for property decide.
     * @param acceptor Index of the property.
     * @return Value of the property at <CODE>index</CODE>.
     */
    public Object getDecide(int acceptor) {
        return decide[acceptor];
    }
    
    /**
     * Getter for property decide.
     * @return Value of property decide.
     */
    public Object[] getDecide() {
        return decide;
    }
    
    /**
     * Indexed setter for property decide.
     * @param acceptor Index of the property.
     * @param value New value of the property at <CODE>index</CODE>.
     */
    public void setDecide(int acceptor, Object value) {
        lockDecide.lock();
            decide[acceptor] = value;
            decideSetted[acceptor] = true;
        lockDecide.unlock();
    }
    
    public boolean isCollected() {
        return collected;
    }
    
    public void collect() {
        collected = true;
    }
    
    public boolean isFrozen() {
        return frozen;
    }
    
    public void freeze() {
        frozen = true;
        addFreeze(me);
    }
    
    public void addFreeze(int acceptor) {
        if(freeze == null) {
            freeze = new TreeSet<Integer>();
        }
        
        freeze.add(acceptor);
    }
    
    public int countFreeze() {
        return freeze.size();
    }
    
    public int countWeak(Object value) {
        lockWeak.lock();
        int res = count(weak,value);
        lockWeak.unlock();
        return res;
    }
    
    public int countStrong(Object value) {
        lockStrong.lock();
        int res = count(strong,value);
        lockStrong.unlock();
        return res;
    }
    
    public int countDecide(Object value) {
        lockDecide.lock();
        int res = count(decide,value);
        lockDecide.unlock();
        return res;
    }
    
    private int count(Object[] array, Object value) {
     
        if(value != null){
            int counter = 0;
            for(int i=0; i < array.length; i++) {
                if(value.equals(array[i])) {
                    counter++;
                }
            }
            return counter;
        }
        
        //br.ufsc.das.util.Logger.println("(Round) count retornou 0 pq value=null");
        return 0;
    
    }
    
    /**
     * Print round information.
     */
    public String toString() {
        StringBuffer buffWeak   = new StringBuffer(1024);
        StringBuffer buffStrong = new StringBuffer(1024);
        StringBuffer buffDecide = new StringBuffer(1024);
        
        buffWeak.append("W=(");
        buffStrong.append("S=(");
        buffDecide.append("D=(");
        
        //recall that weak.length = strong.length = decide.length
        for(int i= 0; i < weak.length-1; i++) {
            buffWeak.append(str(weak[i])+",");
            buffStrong.append(str(strong[i])+",");
            buffDecide.append(str(decide[i])+",");
        }
        
        buffWeak.append(str(weak[weak.length-1])+")");
        buffStrong.append(str(strong[strong.length-1])+")");
        buffDecide.append(str(decide[decide.length-1])+")");
        
        return "eid="+execution.getId()+" r="+getNumber()+" "+buffWeak+" "+buffStrong+" "+buffDecide;
    }
    
    private String str(Object obj) {
        return (obj == null)?"*":obj.toString();
    }
    
}
